#include "snd.h"
#include <X11/cursorfont.h>

enum {menu_menu,
        file_menu,f_cascade_menu,
          f_open_menu,f_close_menu,f_save_menu,f_save_as_menu,f_revert_menu,f_exit_menu,f_new_menu,
          f_view_menu,f_print_menu,f_mix_menu,f_update_menu,f_record_menu,f_sep_menu,
        edit_menu,e_cascade_menu,
          e_cut_menu,e_paste_menu,e_mix_menu,e_play_menu,e_save_as_menu,e_undo_menu,
          e_redo_menu,e_find_menu,e_edenv_menu,e_header_menu,
#if (XmVERSION == 1)
          e_history_menu,
#endif
          e_select_all_menu,
          e_select_sep_menu,e_edit_sep_menu,
        help_menu,h_cascade_menu,
          h_click_for_help_menu,h_about_snd_menu,h_fft_menu,h_find_menu,h_undo_menu,h_sync_menu,h_speed_menu,
          h_expand_menu,h_contrast_menu,h_reverb_menu,h_env_menu,h_marks_menu,h_sound_files_menu,h_init_file_menu,
          h_mix_menu,h_recording_menu,h_clm_menu,h_news_menu,
        option_menu,o_cascade_menu,
          o_transform_menu,
          o_focus_style_menu,o_focus_cascade_menu,
            o_focus_right_menu,o_focus_left_menu,o_focus_middle_menu,o_focus_active_menu,
          o_save_menu,o_save_state_menu,
          o_speed_menu,o_speed_cascade_menu,
            o_speed_float_menu,o_speed_ratio_menu,o_speed_semitone_menu,
          o_stats_menu,
        view_menu,v_cascade_menu,
          v_normalize_menu, 
          v_graph_style_menu, v_graph_style_cascade_menu,
            v_lines_menu,v_dots_menu,v_filled_menu,v_dots_and_lines_menu,v_lollipops_menu,
          v_marks_menu, v_zero_menu, v_axes_menu, v_cursor_menu, v_ctrls_menu, v_listener_menu,
          v_region_menu,
          v_combine_menu, v_combine_cascade_menu,
            v_combine_separate_menu,v_combine_combined_menu,v_combine_superimposed_menu,
          v_color_menu, v_orientation_menu, 
          v_files_menu, v_groups_menu, v_consoles_menu,
          v_x_axis_menu,v_x_axis_cascade_menu,
            v_x_axis_seconds_menu,v_x_axis_samples_menu,v_x_axis_percentage_menu,
          v_error_history_menu,
          v_sep1_menu,v_sep2_menu
};

#if (XmVERSION == 1)
  #define NUM_MENU_WIDGETS 104
#else
  #define NUM_MENU_WIDGETS 103
#endif

static Widget mw[NUM_MENU_WIDGETS];

enum {W_pop_menu,W_pop_sep,W_pop_play,W_pop_undo,W_pop_redo,W_pop_save,W_pop_normalize,W_pop_info};
#define NUM_POPUP_CHILDREN 8
static Widget popup_menu = NULL;
static Widget popup_children[NUM_POPUP_CHILDREN];

void reflect_file_open_in_menu (void)
{
  XtSetSensitive(mw[f_close_menu],TRUE);
  XtSetSensitive(mw[f_print_menu],TRUE);
  XtSetSensitive(mw[f_mix_menu],TRUE);
  XtSetSensitive(mw[f_save_as_menu],TRUE);
  XtSetSensitive(mw[f_update_menu],TRUE);  
  XtSetSensitive(mw[v_normalize_menu],TRUE);  
  XtSetSensitive(mw[e_header_menu],TRUE);
  XtSetSensitive(mw[e_find_menu],TRUE);
  XtSetSensitive(mw[e_select_all_menu],TRUE);
  if (popup_menu)
    {
      XtSetSensitive(popup_children[W_pop_normalize],TRUE);
      XtSetSensitive(popup_children[W_pop_play],TRUE);
      XtSetSensitive(popup_children[W_pop_info],TRUE);
    }
}

void reflect_file_change_in_menu (void)
{
  XtSetSensitive(mw[f_save_menu],TRUE);
  XtSetSensitive(mw[f_revert_menu],TRUE);
  XtSetSensitive(mw[e_undo_menu],TRUE);
  if (popup_menu)
    {
      XtSetSensitive(popup_children[W_pop_undo],TRUE);
      XtSetSensitive(popup_children[W_pop_save],TRUE);
      XtSetSensitive(popup_children[W_pop_play],TRUE);
    }
}

void reflect_file_lack_in_menu (void)
{
  XtSetSensitive(mw[f_close_menu],FALSE);
  XtSetSensitive(mw[f_save_as_menu],FALSE);
  XtSetSensitive(mw[f_save_menu],FALSE);
  XtSetSensitive(mw[f_revert_menu],FALSE);
  XtSetSensitive(mw[f_print_menu],FALSE);
  XtSetSensitive(mw[f_mix_menu],FALSE);
  XtSetSensitive(mw[f_update_menu],FALSE);
  XtSetSensitive(mw[e_undo_menu],FALSE);
  XtSetSensitive(mw[e_redo_menu],FALSE);
  XtSetSensitive(mw[v_normalize_menu],FALSE);
  XtSetSensitive(mw[e_header_menu],FALSE);
  XtSetSensitive(mw[e_find_menu],FALSE);
  XtSetSensitive(mw[e_select_all_menu],FALSE);
  if (popup_menu)
    {
      XtSetSensitive(popup_children[W_pop_undo],FALSE);
      XtSetSensitive(popup_children[W_pop_redo],FALSE);
      XtSetSensitive(popup_children[W_pop_save],FALSE);
      XtSetSensitive(popup_children[W_pop_play],FALSE);
      XtSetSensitive(popup_children[W_pop_info],FALSE);
      XtSetSensitive(popup_children[W_pop_normalize],FALSE);
    }
}

void reflect_mix_active_in_menu(void)
{
  XtSetSensitive(mw[v_consoles_menu],TRUE);
}

void set_normalize_option(int on)
{
  XtSetSensitive(mw[v_normalize_menu],on);
}

static int find_any_edits (chan_info *cp, void *ptr)
{
  return(cp->edit_ctr);
}

static int find_any_possible_edits (chan_info *cp, void *ptr)
{
  return(cp->edit_size);
}

void reflect_file_revert_in_menu (snd_state *ss)
{
  int editing;
  editing = map_over_chans(ss,find_any_edits,NULL);
  if (!editing)
    {
      XtSetSensitive(mw[f_save_menu],FALSE);
      XtSetSensitive(mw[f_revert_menu],FALSE);
      XtSetSensitive(mw[e_undo_menu],FALSE);
      if (popup_menu)
	{
	  XtSetSensitive(popup_children[W_pop_undo],FALSE);
	  XtSetSensitive(popup_children[W_pop_save],FALSE);
	}
    }
  XtSetSensitive(mw[e_redo_menu],TRUE);
  if (popup_menu) XtSetSensitive(popup_children[W_pop_redo],TRUE);
}

void reflect_file_save_in_menu (snd_state *ss)
{
  int editing;
  editing = map_over_chans(ss,find_any_edits,NULL);
  if (!editing)
    {
      XtSetSensitive(mw[f_save_menu],FALSE);
      XtSetSensitive(mw[f_revert_menu],FALSE);
      XtSetSensitive(mw[e_undo_menu],FALSE);
      if (popup_menu)
	{
	  XtSetSensitive(popup_children[W_pop_undo],FALSE);
	  XtSetSensitive(popup_children[W_pop_save],FALSE);
	  XtSetSensitive(popup_children[W_pop_redo],FALSE);
	}
      XtSetSensitive(mw[e_redo_menu],FALSE);
    }
  editing = map_over_chans(ss,find_any_possible_edits,NULL);
}

void reflect_file_revert_in_label (snd_info *sp)
{
  int editing;
  editing = map_over_sound_chans(sp,find_any_edits,NULL);
  if (!editing)
    {
      make_name_label(w_snd_name(sp),shortname(sp));
      make_a_big_star_outa_me(sp->state,sp->shortname,0);
    }
}

void reflect_no_more_redo_in_menu(void)
{
  XtSetSensitive(mw[e_redo_menu],FALSE);
  if (popup_menu) XtSetSensitive(popup_children[W_pop_redo],FALSE);
}

void reflect_edit_with_selection_in_menu (void)
{
  XtSetSensitive(mw[e_cut_menu],TRUE);
  XtSetSensitive(mw[e_paste_menu],TRUE);
  XtSetSensitive(mw[e_play_menu],TRUE);
  XtSetSensitive(mw[e_mix_menu],TRUE);
  XtSetSensitive(mw[e_save_as_menu],TRUE);
  enved_reflect_selection(TRUE);
}

void reflect_edit_without_selection_in_menu (void)
{
  XtSetSensitive(mw[e_cut_menu],FALSE);
  if (!(region_ok(0))) XtSetSensitive(mw[e_paste_menu],FALSE);
  XtSetSensitive(mw[e_play_menu],FALSE);
  enved_reflect_selection(FALSE);
}

void reflect_undo_in_menu(void)
{
  XtSetSensitive(mw[e_redo_menu],TRUE);
  if (popup_menu) XtSetSensitive(popup_children[W_pop_redo],TRUE);
}

void reflect_redo_in_menu(void)
{
  XtSetSensitive(mw[e_undo_menu],TRUE);
  if (popup_menu) XtSetSensitive(popup_children[W_pop_undo],TRUE);
  reflect_file_change_in_menu();
}

void reflect_undo_ok_in_menu(void)
{
  XtSetSensitive(mw[e_undo_menu],TRUE);
  if (popup_menu) XtSetSensitive(popup_children[W_pop_undo],TRUE);
}

void reflect_undo_or_redo_in_menu(chan_info *cp)
{
  int undoable,redoable;
  if (cp)
    {
      undoable = (cp->edit_ctr > 0);
      redoable = (!(((cp->edit_ctr+1) == cp->edit_size) || (!(cp->edits[cp->edit_ctr+1]))));
      XtSetSensitive(mw[e_undo_menu],undoable);
      if (popup_menu) XtSetSensitive(popup_children[W_pop_undo],undoable);
      XtSetSensitive(mw[e_redo_menu],redoable);
      if (popup_menu) XtSetSensitive(popup_children[W_pop_redo],redoable);
    }
}

void reflect_regions_in_menu(void)
{
  XtSetSensitive(mw[v_region_menu],TRUE);
  XtSetSensitive(mw[e_save_as_menu],TRUE);
}

void reflect_no_regions_in_menu(void)
{
  XtSetSensitive(mw[v_region_menu],FALSE);
  XtSetSensitive(mw[e_save_as_menu],FALSE);
}

void reflect_raw_open_in_menu(void)
{
  XtSetSensitive(mw[f_open_menu],TRUE);
  XtSetSensitive(mw[f_view_menu],TRUE);
  XtSetSensitive(mw[f_new_menu],TRUE);
}

void reflect_raw_pending_in_menu(void)
{
  XtSetSensitive(mw[f_open_menu],FALSE);
  XtSetSensitive(mw[f_view_menu],FALSE);
  XtSetSensitive(mw[f_new_menu],FALSE);
}


/* -------------------------------- FILE MENU -------------------------------- */

static void File_Open_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_open_file_dialog(w,(snd_state *)clientData);
}

static void File_View_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  ss->viewing = 1;
  snd_open_file_dialog(w,ss);
}

static void File_New_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_new_file_checked((snd_state *)clientData,NULL,unsupported_sound_file,SNDLIB_UNSUPPORTED,0,0);
}

static void File_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"File Menu","#fileoperations");
#else
  ssnd_help((snd_state *)clientData,
          "File Menu",
"The File menu provides one way to open,\n\
close, and save files. Its options are:\n\
\n",
get_file_menu_help(),
"\n\
The Print option produces a Postscript file;\n\
you can print it with lpr.\n",
NULL);
#endif
}

static void File_Record_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_record_file((snd_state *)clientData);
}

static void File_Close_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  snd_info *sp;
  finish_keyboard_selection();
  sp = any_selected_sound(ss);
  if (sp) snd_close_file(sp,ss);
}

static void File_Save_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_info *sp;
  snd_state *ss = (snd_state *)clientData;
  finish_keyboard_selection();
  sp = any_selected_sound(ss);
  save_edits(sp,NULL);
}

static int file_update(snd_info *sp, void *ptr)
{
  snd_state *ss = (snd_state *)ptr;
  if ((sp) && (sp->edited_region == NULL)) snd_update(ss,sp);
  return(0);
}

static void File_Update_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  finish_keyboard_selection();
  map_over_sounds(ss,file_update,(void *)ss);
}

static void File_Save_As_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_save_as_dialog(w,(snd_state *)clientData);
}

static void File_Revert_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  snd_info *sp;
  int i;
  finish_keyboard_selection();
  sp = any_selected_sound(ss);
  for (i=0;i<sp->nchans;i++) revert_edits(sp->chans[i],NULL);
  reflect_file_revert_in_label(sp);
  reflect_file_revert_in_menu(ss);
}

static void File_Exit_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  finish_keyboard_selection();
#if HAVE_GUILE
  if (dont_exit(ss)) return;
#endif
  snd_exit_cleanly(ss);
  snd_exit(1);
}



/* -------------------------------- EDIT MENU -------------------------------- */


static void Edit_Mix_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  chan_info *cp;
  finish_keyboard_selection();
  cp = selected_channel((snd_state *)clientData);
  if (cp) add_region(0,cp,"Edit: mix");
}

static void Edit_Envelope_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  create_envelope_editor((snd_state *)clientData);
}

static void Edit_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"Edit Menu","#editoperations");
#else
  ssnd_help((snd_state *)clientData,
	    "Edit Menu",
"The Edit Menu options apply to the\n\
current selection in most cases.  The\n\
successive selections are saved on a stack\n\
of 'regions' accessible via ctrl-Y.\n\
The selection can be retrieved by or\n\
from other programs or within Snd with\n\
the middle mouse button.  The options are:\n\
\n",
get_edit_menu_help(),
NULL);
#endif
}

static void Edit_Cut_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  finish_keyboard_selection();
  if (region_ok(0))
    {
      delete_selection("Edit: Cut",UPDATE_DISPLAY);
    }
}

static void Edit_Paste_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  chan_info *cp;
  finish_keyboard_selection();
  cp = selected_channel((snd_state *)clientData);
  if (cp) paste_region(0,cp,"Edit: Paste");
}

static void Edit_Save_As_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  /* refers to the selection */
  snd_edit_save_as_dialog(w,(snd_state *)clientData);
}

static void Edit_Select_All_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  chan_info *cp;
  cp = current_channel(ss);
  if (cp) define_region(cp,0,current_ed_samples(cp),TRUE);
  map_over_chans(ss,update_graph,NULL);
}

static void Edit_Undo_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  finish_keyboard_selection();
  undo_EDIT((void *)clientData,1);
}

static void Edit_Redo_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  finish_keyboard_selection();
  redo_EDIT((void *)clientData,1);
}

static void Edit_Header_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  snd_info *sp;
  sp = selected_sound(ss);
  if (sp) edit_header(sp);
}

static void Edit_Play_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  finish_keyboard_selection();
  if (region_ok(0)) play_region((snd_state *)clientData,0,NULL);
}

#if (XmVERSION == 1)
void edit_history(snd_state *ss, int on)
{
  /* update menu label also */
  int wid;
  if ((on) && (edit_history_width(ss) == 0))
    {
      wid = get_window_width(main_PANE(ss));
      if (wid > 200) wid = 100; else wid = wid/2;
      in_set_edit_history_width(ss,wid); 
    }
  set_show_edit_history(ss,on);
  make_name_label(mw[e_history_menu],(on) ? STR_Hide_edit_history : STR_Show_edit_history);
  if (on)
    map_over_chans(ss,open_edit_histories,NULL);
  else map_over_chans(ss,close_edit_histories,NULL);
}

static void Edit_History_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  edit_history(ss,(!(show_edit_history(ss))));
}
#endif


/* -------------------------------- VIEW MENU -------------------------------- */

static void View_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"View Menu","#viewing");
#else
  ssnd_help((snd_state *)clientData,
	    "View Menu",
"The View Menu affects the overall Snd display.\n\
Its options are:\n\
\n\
",
get_view_menu_help(),
NULL);
#endif
}

static int update_axes(chan_info *cp, void *ptr)
{
  set_xy_bounds(cp,cp->axis);
  return(0);
}

void set_wavo(snd_state *ss, int on)
{
  in_set_wavo(ss,on);
  if (on == 0) map_over_chans(ss,update_axes,NULL);
  map_over_chans(ss,update_graph,NULL);
}

static int update_sound(snd_info *sp, void *ptr)
{
  snd_state *ss = (snd_state *)ptr;
  if (sp)
    {
      switch (channel_style(ss))
	{
	case CHANNELS_SEPARATE: separate_sound(sp); break;
	case CHANNELS_COMBINED: combine_sound(sp); break;
	case CHANNELS_SUPERIMPOSED: superimpose_sound(sp); break;
	}
    }
  return(0);
}

void set_channel_style(snd_state *ss, int val)
{
  switch (channel_style(ss))
    {
    case CHANNELS_SEPARATE: XtSetSensitive(mw[v_combine_separate_menu],TRUE); break;
    case CHANNELS_COMBINED: XtSetSensitive(mw[v_combine_combined_menu],TRUE); break;
    case CHANNELS_SUPERIMPOSED: XtSetSensitive(mw[v_combine_superimposed_menu],TRUE); break;
    }
  in_set_channel_style(ss,val);
  switch (val)
    {
    case CHANNELS_SEPARATE: XtSetSensitive(mw[v_combine_separate_menu],FALSE); break;
    case CHANNELS_COMBINED: XtSetSensitive(mw[v_combine_combined_menu],FALSE); break;
    case CHANNELS_SUPERIMPOSED: XtSetSensitive(mw[v_combine_superimposed_menu],FALSE); break;
    }
  map_over_sounds(ss,update_sound,(void *)ss);
  map_over_chans(ss,update_graph,NULL);
}

static void View_Separate_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_channel_style(ss,CHANNELS_SEPARATE);
}

static void View_Combined_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_channel_style(ss,CHANNELS_COMBINED);
}

static void View_Superimposed_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_channel_style(ss,CHANNELS_SUPERIMPOSED);
}


static void View_Normalize_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  normalize_all_sounds((snd_state *)clientData);
}


void set_graph_style(snd_state *ss, int val)
{
  switch (graph_style(ss))
    {
    case GRAPH_LINES: XtSetSensitive(mw[v_lines_menu],TRUE); break;
    case GRAPH_DOTS: XtSetSensitive(mw[v_dots_menu],TRUE); break;
    case GRAPH_FILLED: XtSetSensitive(mw[v_filled_menu],TRUE); break;
    case GRAPH_DOTS_AND_LINES: XtSetSensitive(mw[v_dots_and_lines_menu],TRUE); break;
    case GRAPH_LOLLIPOPS: XtSetSensitive(mw[v_lollipops_menu],TRUE); break;
    }
  in_set_graph_style(ss,val);
  map_over_chans(ss,update_graph,NULL);
  switch (val)
    {
    case GRAPH_LINES: XtSetSensitive(mw[v_lines_menu],FALSE); break;
    case GRAPH_DOTS: XtSetSensitive(mw[v_dots_menu],FALSE); break;
    case GRAPH_FILLED: XtSetSensitive(mw[v_filled_menu],FALSE); break;
    case GRAPH_DOTS_AND_LINES: XtSetSensitive(mw[v_dots_and_lines_menu],FALSE); break;
    case GRAPH_LOLLIPOPS: XtSetSensitive(mw[v_lollipops_menu],FALSE); break;
    }
}

static void View_Dots_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_graph_style(ss,GRAPH_DOTS);
}

static void View_Lines_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_graph_style(ss,GRAPH_LINES);
}

static void View_Filled_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_graph_style(ss,GRAPH_FILLED);
}

static void View_Dots_and_Lines_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_graph_style(ss,GRAPH_DOTS_AND_LINES);
}

static void View_Lollipops_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_graph_style(ss,GRAPH_LOLLIPOPS);
}

void set_show_marks(snd_state *ss, int val)
{
  in_set_show_marks(ss,val);
  if (mw[v_marks_menu])
    {
      make_name_label(mw[v_marks_menu],(val) ? STR_Hide_marks : STR_Show_marks);
      map_over_chans(ss,update_graph,NULL);
    }
}

static void View_Marks_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  /* similar to dots menus */
  snd_state *ss = (snd_state *)clientData;
  set_show_marks(ss,(!(show_marks(ss))));
}

void set_show_y_zero(snd_state *ss, int val)
{
  in_set_show_y_zero(ss,val);
  if (mw[v_zero_menu])
    {
      make_name_label(mw[v_zero_menu],(val) ? STR_Hide_Y0 : STR_Show_Y0);
      map_over_chans(ss,update_graph,NULL);
    }
}

static void View_Zero_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_show_y_zero(ss,(!(show_y_zero(ss))));
}

void set_show_axes(snd_state *ss, int val)
{
  in_set_show_axes(ss,val);
  if (mw[v_axes_menu])
    {
      make_name_label(mw[v_axes_menu],(val) ? STR_Hide_axes : STR_Show_axes);
      map_over_chans(ss,update_graph,NULL);
    }
}

static void View_Axes_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_show_axes(ss,(!(show_axes(ss))));
}

static int clrmini(snd_info *sp, void *ptr) {clear_minibuffer(sp); return(0);}

void set_verbose_cursor(snd_state *ss, int val)
{
  in_set_verbose_cursor(ss,val);
  if (val == 0) map_over_sounds(ss,clrmini,NULL);
  if (mw[v_cursor_menu])
	make_name_label(mw[v_cursor_menu],(val) ? STR_Silent_cursor : STR_Verbose_cursor);
}

static void View_Cursor_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_verbose_cursor(ss,(!(verbose_cursor(ss))));
}

void set_view_ctrls_label(char *lab)
{
  make_name_label(mw[v_ctrls_menu],lab);
}

static void View_Ctrls_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  if (ss->ctrls_height < 100) show_controls(ss); else hide_controls(ss); /* snd-xmain.c */
}

void set_view_listener_label(char *lab)
{
  make_name_label(mw[v_listener_menu],lab);
}

static void View_Listener_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  handle_listener(ss,LISTENER_OPEN);
}

void set_show_mix_consoles(snd_state *ss, int on)
{
  in_set_show_mix_consoles(ss,on);
  make_name_label(mw[v_consoles_menu],(on) ? STR_Hide_consoles : STR_Show_consoles);
  update_all_consoles(ss);
}

static void View_Groups_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  fire_up_group_browser((snd_state *)clientData);
}

static void View_Consoles_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_show_mix_consoles(ss,!(show_mix_consoles(ss)));
}

static void View_Error_History_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  show_snd_errors((snd_state *)clientData);
}



/* -------------------------------- OPTIONS MENU -------------------------------- */

static void Options_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"Options Menu","#options");
#else
  ssnd_help((snd_state *)clientData,
	   "Options Menu",
"The Options menu items affect how the FFT\n\
operates, and whatnot.  The items are:\n\
\n\
",
get_options_menu_help(),
NULL);
#endif
}

static void Options_Transform_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  fire_up_transform_dialog((snd_state *)clientData);
}


/* Save Options */

static void Options_Save_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  char *buf = NULL;
  if (save_options(ss) == 0)
    {
      if (any_selected_sound(ss))
	{
	  buf = (char *)CALLOC(128,sizeof(char));
	  sprintf(buf,"saved options in %s",ss->init_file);
	  report_in_minibuffer(any_selected_sound(ss),buf);
	  FREE(buf);
	}
    }
  else 
    snd_error(STR_cant_write_p,ss->init_file,strerror(errno));
}

void activate_focus_menu(snd_state *ss, int new_focus)
{
  if (mw[o_focus_left_menu])
    {
      switch (zoom_focus_style(ss))
	{
	case FOCUS_LEFT: XtSetSensitive(mw[o_focus_left_menu],TRUE); break;
	case FOCUS_RIGHT: XtSetSensitive(mw[o_focus_right_menu],TRUE); break;
	case FOCUS_MIDDLE: XtSetSensitive(mw[o_focus_middle_menu],TRUE); break;
	case FOCUS_ACTIVE: XtSetSensitive(mw[o_focus_active_menu],TRUE); break;
	}
    }
  set_zoom_focus_style(ss,new_focus);
  if (mw[o_focus_left_menu])
    {
      switch (new_focus)
	{
	case FOCUS_LEFT: XtSetSensitive(mw[o_focus_left_menu],FALSE); break;
	case FOCUS_RIGHT: XtSetSensitive(mw[o_focus_right_menu],FALSE); break;
	case FOCUS_MIDDLE: XtSetSensitive(mw[o_focus_middle_menu],FALSE); break;
	case FOCUS_ACTIVE: XtSetSensitive(mw[o_focus_active_menu],FALSE); break;
	}
    }
}  

static void Options_Focus_Right_Callback(Widget w,XtPointer clientData,XtPointer Data) {activate_focus_menu((snd_state *)clientData,FOCUS_RIGHT);}
static void Options_Focus_Left_Callback(Widget w,XtPointer clientData,XtPointer Data) {activate_focus_menu((snd_state *)clientData,FOCUS_LEFT);}
static void Options_Focus_Middle_Callback(Widget w,XtPointer clientData,XtPointer Data) {activate_focus_menu((snd_state *)clientData,FOCUS_MIDDLE);}
static void Options_Focus_Active_Callback(Widget w,XtPointer clientData,XtPointer Data) {activate_focus_menu((snd_state *)clientData,FOCUS_ACTIVE);}


void activate_speed_in_menu(snd_state *ss, int newval)
{
  if (mw[o_speed_ratio_menu])
    {
      switch (speed_style(ss))
	{
	case SPEED_AS_RATIO: XtSetSensitive(mw[o_speed_ratio_menu],TRUE); break;
	case SPEED_AS_SEMITONE: XtSetSensitive(mw[o_speed_semitone_menu],TRUE); break;
	default: XtSetSensitive(mw[o_speed_float_menu],TRUE); break;
	}
    }
  in_set_speed_style(ss,newval);
  if (mw[o_speed_ratio_menu])
    {
      switch (speed_style(ss))
	{
	case SPEED_AS_RATIO: XtSetSensitive(mw[o_speed_ratio_menu],FALSE); break;
	case SPEED_AS_SEMITONE: XtSetSensitive(mw[o_speed_semitone_menu],FALSE); break;
	default: XtSetSensitive(mw[o_speed_float_menu],FALSE); break;
	}
    }
}

static void Options_Speed_Float_Callback(Widget w,XtPointer cD,XtPointer Data) {activate_speed_in_menu((snd_state *)cD,SPEED_AS_FLOAT);}
static void Options_Speed_Ratio_Callback(Widget w,XtPointer cD,XtPointer Data) {activate_speed_in_menu((snd_state *)cD,SPEED_AS_RATIO);}
static void Options_Speed_Semitone_Callback(Widget w,XtPointer cD,XtPointer Data) {activate_speed_in_menu((snd_state *)cD,SPEED_AS_SEMITONE);}

static void reflect_x_axis_unit_change_in_menu(int oldval, int newval)
{
  switch (oldval)
    {
    case X_IN_SECONDS: XtSetSensitive(mw[v_x_axis_seconds_menu],TRUE); break;
    case X_IN_SAMPLES: XtSetSensitive(mw[v_x_axis_samples_menu],TRUE); break;
    case X_TO_ONE: XtSetSensitive(mw[v_x_axis_percentage_menu],TRUE); break;
    default: XtSetSensitive(mw[v_x_axis_seconds_menu],TRUE); break;
    }
  switch (newval)
    {
    case X_IN_SECONDS: XtSetSensitive(mw[v_x_axis_seconds_menu],FALSE); break;
    case X_IN_SAMPLES: XtSetSensitive(mw[v_x_axis_samples_menu],FALSE); break;
    case X_TO_ONE: XtSetSensitive(mw[v_x_axis_percentage_menu],FALSE); break;
    default: XtSetSensitive(mw[v_x_axis_seconds_menu],FALSE); break;
    }
}
  
void set_x_axis_style(snd_state *ss, int val)
{
  reflect_x_axis_unit_change_in_menu(x_axis_style(ss),val);
  in_set_x_axis_style(ss,val);
  map_over_chans(ss,update_graph,NULL);
}

static void Options_X_Axis_Seconds_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  set_x_axis_style((snd_state *)clientData,X_IN_SECONDS);
}

static void Options_X_Axis_Samples_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  set_x_axis_style((snd_state *)clientData,X_IN_SAMPLES);
}

static void Options_X_Axis_Percentage_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  set_x_axis_style((snd_state *)clientData,X_TO_ONE);
}

void set_save_state_file(snd_state *ss, char *name)
{
  in_set_save_state_file(ss,name);
  XtSetSensitive(mw[o_save_state_menu],(snd_strlen(name) > 0));
}

static void Options_Save_State_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  char *buf = NULL;
  if (save_state_file(ss))
    {
      if (save_state(ss,save_state_file(ss)) == 0)
	{
	  if (any_selected_sound(ss))
	    {
	      buf = (char *)CALLOC(128,sizeof(char));
	      sprintf(buf,"saved state in %s",save_state_file(ss));
	      report_in_minibuffer(any_selected_sound(ss),buf);
	      FREE(buf);
	    }
	}
    }
}

void set_show_usage_stats(snd_state *ss, int val)
{
  in_set_show_usage_stats(ss,val);
  if (mw[o_stats_menu])
    make_name_label(mw[o_stats_menu],(val) ? STR_Ignore_stats : STR_Show_stats);
  check_stats_window(ss,val);
}

static void Options_Stats_Callback(Widget w,XtPointer clientData,XtPointer callData)
{
  snd_state *ss = (snd_state *)clientData;
  set_show_usage_stats(ss,(!(show_usage_stats(ss))));
}




/* -------------------------------- HELP MENU -------------------------------- */

static void Help_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"Help Menu","#menus");
#else
  ssnd_help((snd_state *)clientData,
          "Help Menu",
"The Help menu tries to provide quick\n\
help for the most common Snd operations.\n\
The menu items are:\n\
\n",
get_help_menu_help(),
NULL);
#endif
}

static void Help_Context_Help_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *)callData;
  Widget selectedWidget = NULL;
  static Cursor cursor = 0;
  Widget parent = XtParent(w);
  snd_state *ss = (snd_state *)clientData;
  Widget mainWindow;
  mainWindow = main_PANE(ss);
  if (!cursor) cursor = XCreateFontCursor(XtDisplay(parent),XC_question_arrow); 
  selectedWidget = XmTrackingLocate(mainWindow,cursor,FALSE);
  if (selectedWidget)
    {
      XmAnyCallbackStruct cb;
      cb.reason = XmCR_HELP;
      cb.event = cbs->event;
      do
        {
	  if ((XtHasCallbacks(selectedWidget,XmNhelpCallback) == XtCallbackHasSome))
            {
	      XtCallCallbacks(selectedWidget,XmNhelpCallback,&cb);
	      return;
            }
	  else
	    selectedWidget = XtParent(selectedWidget);
        } while (selectedWidget != NULL);
    }
}

static void Help_About_Snd_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,"Snd","#gettingstarted");
#else
  ssnd_help((snd_state *)clientData,
               "Snd",
get_about_snd_help(),
"The various Help menu items are:\n\
\n",
get_help_menu_help(),
"\n",
"The main menu items are:\n\
  File: operations on files.\n\
  Edit: operations on the current selection.\n\
  View: change Snd display choices.\n\
  Options: change Snd analysis choices.\n\
  Help: this menu.\n\
The main menu bar itself is Snd's 'drop' box.\n\
\n\
The File menu's options are:\n\
",
get_file_menu_help(),
"\n\
The Edit menu's options are:\n\
",
get_edit_menu_help(),
"\n\
The View menu's options are:\n\
",
get_view_menu_help(),
"\n\
The Options menu's items are:\n\
",
get_options_menu_help(),
"\n\
The graph editing commands are:\n\
",
get_graph_help(),
"\n\
",
get_edit_history_help(),
"\n\
",
get_fft_keypad_help(),
"\n\
",
get_mark_help(),
"\n\
",
get_init_file_help(),
"\n\
",
get_resource_help(),
"\n\
",
get_expression_help(),
NULL);
#endif
}

static void Help_FFT_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_FFT,"#viewfft");
#else
  ssnd_help((snd_state *)clientData,
	    STR_FFT,
get_fft_menu_help(),
"\n\
",
get_fft_keypad_help(),
NULL);
#endif
}

static void Help_Find_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Find,"#find");
#else
  ssnd_help((snd_state *)clientData,
	    STR_Find,
get_find_menu_help(),
"\n\
",
get_expression_help(),
NULL);
#endif
}

static void Help_Undo_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Undo,"#undoredo");
#else
  snd_help((snd_state *)clientData,STR_Undo,get_undo_menu_help());
#endif
}

static void Help_Sync_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Sync,"#multichannel");
#else
  snd_help((snd_state *)clientData,STR_Sync,get_sync_menu_help());
#endif
}

static void Help_Speed_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Speed,"#speed");
#else
  snd_help((snd_state *)clientData,STR_Speed,get_speed_menu_help());
#endif
}

static void Help_Expand_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Expand,"#expand");
#else
  snd_help((snd_state *)clientData,STR_Expand,get_expand_menu_help());
#endif
}

static void Help_Reverb_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Reverb,"#reverb");
#else
  snd_help((snd_state *)clientData,STR_Reverb,get_reverb_menu_help());
#endif
}

static void Help_Contrast_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Contrast,"#contrast");
#else
  snd_help((snd_state *)clientData,STR_Contrast,get_contrast_menu_help());
#endif
}

static void Help_Env_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Envelope,"#editenvelope");
#else
  ssnd_help((snd_state *)clientData,
	    STR_Envelope,
get_env_menu_help(),
"\n\
The simplest way to define and apply envelopes\n\
is to use the envelope editor via the Edit menu.\n\
\n\
",
get_envelope_editor_help(),
NULL);
#endif
}

static void Help_Marks_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Marks,"#marks");
#else
  snd_help((snd_state *)clientData,STR_Marks,get_mark_help());
#endif
}

static void Help_Mix_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Mixing,"#mixingfiles");
#else
  ssnd_help((snd_state *)clientData,STR_Mixing,get_mixing_help(),"\n",get_grouping_help(),NULL);
#endif
}

static void Help_Sound_Files_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Format,"#formats");
#else
  snd_help((snd_state *)clientData,STR_Format,get_format_menu_help());
#endif
}

static void Help_Init_File_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Customization,"extsnd.html");
#else
  ssnd_help((snd_state *)clientData,STR_Customization,get_init_file_help(),"\n",get_resource_help(),NULL);
#endif
}

static void Help_Recording_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_Recording,"#recordfile");
#else
  snd_help((snd_state *)clientData,STR_Recording,get_recording_help());
#endif
}

#if WITH_MUS_MODULE
char *get_CLM_help(void);
static void Help_CLM_Callback (Widget w,XtPointer clientData,XtPointer callData)
{
#if HAVE_XmHTML
  snd_help((snd_state *)clientData,STR_CLM,"extsnd.html#sndwithclm");
#else
  snd_help((snd_state *)clientData,STR_CLM,get_CLM_help());
#endif
}
#endif

static char *snd_itoa(int n)
{
  char *str;
  str = (char *)CALLOC(8,sizeof(char));
  sprintf(str,"%d",n);
  return(str);
}

#if HAVE_XPM
  #include <X11/xpm.h>
#endif
#if HAVE_XmHTML
  #include <XmHTML/XmHTML.h>
#endif
#if WITH_MUS_MODULE
  #include "clm.h"
#endif
#if 0
#if defined(__GNUC__) && defined(LINUX)
  #include <gnu/libc-version.h>
#endif
#endif

#include <X11/IntrinsicP.h>

static char* vstrcat(char *buf,...)
{
  va_list ap;
  char *str;
  va_start(ap,buf);
  while ((str = va_arg(ap,char *)))
    {
      strcat(buf,str);
    }
  va_end(ap);
  return(buf);
}

char *version_info(void)
{
  #define NUM_ITOAS 17
  char *buf;
  char **itoa;
  int i;
  buf = (char *)CALLOC(1024,sizeof(char));
  itoa = (char **)CALLOC(NUM_ITOAS,sizeof(char *));
  vstrcat(buf,
	  "This is Snd version ",
	  SND_RPM_VERSION,
	  " of ",
	  SND_VERSION,
	  ": ",
#ifdef SND_AS_WIDGET
	  "\n    (compiled as a widget)",
#endif
#if HAVE_GUILE
	  "\n    Guile ",guile_version(),
#endif
#ifdef WITH_MUS_MODULE
	  "\n    CLM ",itoa[0]=snd_itoa(MUS_VERSION),".",itoa[1]=snd_itoa(MUS_REVISION)," (",MUS_DATE,")",
#endif
#if HAVE_XPM
	  "\n    Xpm ",itoa[2]=snd_itoa(XpmFormat),".",itoa[3]=snd_itoa(XpmVersion),".",itoa[4]=snd_itoa(XpmRevision),
#endif
#ifdef SND_CONF
	  "\n    conf: ",SND_CONF,
#endif
#if HAVE_XmHTML
	  "\n    XmHTML ",itoa[5]=snd_itoa(XmHTMLVERSION),".",itoa[6]=snd_itoa(XmHTMLREVISION),".",itoa[7]=snd_itoa(XmHTMLUPDATE_LEVEL),
#endif
#ifdef LESSTIF_VERSION
	  "\n    Lesstif ",itoa[8]=snd_itoa(LESSTIF_VERSION),".",itoa[9]=snd_itoa(LESSTIF_REVISION)," ",
#endif
	  "\n    Motif ",itoa[10]=snd_itoa(XmVERSION),".",itoa[11]=snd_itoa(XmREVISION),".",itoa[12]=snd_itoa(XmUPDATE_LEVEL)," X",itoa[13]=snd_itoa(X_PROTOCOL),"R",itoa[14]=snd_itoa(XT_REVISION),
#ifdef WITH_BIG_COLORMAP
	  ", (with big colormaps)",
#endif
#ifdef CCRMA
	  "\n    (uses ccrma-specific /zap dirs)",
#endif
	  "\n    ",audio_moniker(),
	  "\n    Sndlib ",itoa[15]=snd_itoa(SNDLIB_VERSION),".",itoa[16]=snd_itoa(SNDLIB_REVISION)," (",SNDLIB_DATE,")",
#ifdef __DATE__
	  "\n    Compiled ",__DATE__," ",__TIME__,
#endif
#ifdef __VERSION__
	  "\n    C: ",__VERSION__,
#endif
#if 0
#if defined(__GNUC__) && defined(LINUX)
	  "\n    Libc: ",gnu_get_libc_version(),".",gnu_get_libc_release(),
#endif
#endif
	  "\n",NULL);
  for (i=0;i<NUM_ITOAS;i++) if (itoa[i]) FREE(itoa[i]);
  FREE(itoa);
  return(buf);
}

static void Help_News_Callback (Widget w,XtPointer clientData,XtPointer callData) 
{
  char *info;
  info = version_info();
  ssnd_help((snd_state *)clientData,STR_News,
	    info,
	    "\n",
	    "Recent changes include:\n\
\n\
8-Nov:   snd-3-4.rpm.\n\
5-Nov:   memory leaks plugged?\n\
         marks bugfix (add mark, save edit, add mark -> segfault).\n\
1-Nov:   start-playing-hook, enved-power.\n\
         Error History Edit menu option.\n\
         if Motif 2, edit history portion is now a horizontal pane of the channel window,\n\
           so Edit History menu option is removed.\n\
28-Oct:  listener-prompt, graph-cursor.\n\
25-Oct:  'env' is 'env-sound' and 'snd-env' is deleted.\n\
         Similarly, convolve -> convolve-arrays,\n\
         filter -> filter-sound,\n\
         src -> src-sound (and snd-src deleted),\n\
         spectrum -> snd-spectrum.\n\
         These changes to avoid confusion with the CLM module.\n\
22-Oct:  audio-output-device.\n\
         sync is now an integer, allowing complex grouping.\n\
15-Oct:  changed variable 'spectro-color' to 'colormap'.\n\
11-Oct:  changed axis drawer so the x axis doesn't jump around.\n\
8-Oct:   dsp-devices (Linux OSS) for setting audio ports by hand, bomb function.\n\
         hook handlers are now taken from Guile (version 1.3.4) and use Scheme functions, rather than strings.\n\
7-Oct:   count-matches (and bind-key) arg can now be a Scheme function, as in find.\n\
         load-colormap.\n\
6-Oct:   play can handle any number of channels (will fold them into available DAC chans).\n\
",
NULL);
  FREE(info);
}



/* -------------------------------- MAIN MENU -------------------------------- */

Widget get_menubar(void) {return(mw[menu_menu]);}

Widget add_menu(snd_state *state)
{
  static Arg main_args[12];
  static Arg in_args[12];
  static Arg high_args[12];
  Arg sep_args[12];
  int in_n,n,high_n,main_n,start_high_n,k,j,p;
  /* this mainly passes the global data pointer (state) to all the menu-related callbacks */
  
  in_n=0;
  main_n=0;
  high_n=0;
  if (!(state->using_schemes))
    {
      main_n = background_basic_color(main_args,main_n,state);
      high_n = background_highlight_color(high_args,high_n,state);
      in_n = background_basic_color(in_args,in_n,state);
    }
  start_high_n = high_n;
  XtSetArg(in_args[in_n],XmNsensitive,FALSE); in_n++;
  
  n = high_n;
  XtSetArg(high_args[n],XmNuserData,state); n++; /* used in snd-xclip by drop site to get main state data from whatever widget got the drop */
  p = n;
  XtSetArg(high_args[p],XmNtopAttachment,XmATTACH_FORM); p++;
  XtSetArg(high_args[p],XmNbottomAttachment,XmATTACH_NONE); p++;
  XtSetArg(high_args[p],XmNleftAttachment,XmATTACH_FORM); p++;
  XtSetArg(high_args[p],XmNrightAttachment,XmATTACH_FORM); p++;
#ifdef SND_AS_WIDGET
  mw[menu_menu] = XtCreateWidget("mb",xmRowColumnWidgetClass,main_PANE(state),high_args,p);
#else
  mw[menu_menu] = XmCreateMenuBar(main_PANE(state),"menuBar",high_args,p);
#endif

  /* FILE MENU */
  mw[file_menu] = XmCreatePulldownMenu(mw[menu_menu],"filem",main_args,main_n);
  
  high_n = start_high_n;
  XtSetArg(high_args[high_n],XmNsubMenuId,mw[file_menu]); high_n++;
  XtSetArg(high_args[high_n],XmNmnemonic,'F'); high_n++;
  mw[f_cascade_menu] = XtCreateManagedWidget(STR_File,xmCascadeButtonWidgetClass,mw[menu_menu],high_args,high_n);
  XtAddCallback(mw[f_cascade_menu],XmNhelpCallback,File_Help_Callback,state);

  mw[f_open_menu] = XtCreateManagedWidget(STR_Open,xmPushButtonWidgetClass,mw[file_menu],main_args,main_n);
  XtAddCallback(mw[f_open_menu],XmNactivateCallback,File_Open_Callback,state);
  XtVaSetValues(mw[f_open_menu],XmNmnemonic,'O',NULL);

  mw[f_close_menu] = XtCreateManagedWidget(STR_Close,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_close_menu],XmNactivateCallback,File_Close_Callback,state);
  XtVaSetValues(mw[f_close_menu],XmNmnemonic,'C',NULL);
  
  mw[f_save_menu] = XtCreateManagedWidget(STR_Save,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_save_menu],XmNactivateCallback,File_Save_Callback,state);
  XtVaSetValues(mw[f_save_menu],XmNmnemonic,'S',NULL);
  
  mw[f_save_as_menu] = XtCreateManagedWidget(STR_Save_as,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_save_as_menu],XmNactivateCallback,File_Save_As_Callback,state);
  XtVaSetValues(mw[f_save_as_menu],XmNmnemonic,'a',NULL);
  
  mw[f_revert_menu] = XtCreateManagedWidget(STR_Revert,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_revert_menu],XmNactivateCallback,File_Revert_Callback,state);
  XtVaSetValues(mw[f_revert_menu],XmNmnemonic,'R',NULL);
  
  mw[f_mix_menu] = XtCreateManagedWidget(STR_Mix,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_mix_menu],XmNactivateCallback,File_Mix_Callback,state);
  XtVaSetValues(mw[f_mix_menu],XmNmnemonic,'M',NULL);

  mw[f_update_menu] = XtCreateManagedWidget(STR_Update,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_update_menu],XmNactivateCallback,File_Update_Callback,state);
  XtVaSetValues(mw[f_update_menu],XmNmnemonic,'U',NULL);

  mw[f_new_menu] = XtCreateManagedWidget(STR_New,xmPushButtonWidgetClass,mw[file_menu],main_args,main_n);
  XtAddCallback(mw[f_new_menu],XmNactivateCallback,File_New_Callback,state);
  XtVaSetValues(mw[f_new_menu],XmNmnemonic,'N',NULL);

  mw[f_record_menu] = XtCreateManagedWidget(STR_Record,xmPushButtonWidgetClass,mw[file_menu],main_args,main_n);
  XtAddCallback(mw[f_record_menu],XmNactivateCallback,File_Record_Callback,state);

  mw[f_view_menu] = XtCreateManagedWidget(STR_View,xmPushButtonWidgetClass,mw[file_menu],main_args,main_n);
  XtAddCallback(mw[f_view_menu],XmNactivateCallback,File_View_Callback,state);
  XtVaSetValues(mw[f_view_menu],XmNmnemonic,'V',NULL);

  mw[f_print_menu] = XtCreateManagedWidget(STR_Print,xmPushButtonWidgetClass,mw[file_menu],in_args,in_n);
  XtAddCallback(mw[f_print_menu],XmNactivateCallback,File_Print_Callback,state);
  XtVaSetValues(mw[f_print_menu],XmNmnemonic,'P',NULL);

  j=0;
  if (!(state->using_schemes)) j = background_basic_color(sep_args,j,state);
  XtSetArg(sep_args[j],XmNseparatorType,XmSHADOW_ETCHED_IN); j++;
  mw[f_sep_menu] = XtCreateManagedWidget("",xmSeparatorWidgetClass,mw[file_menu],sep_args,j);

  mw[f_exit_menu] = XtCreateManagedWidget(STR_Exit,xmPushButtonWidgetClass,mw[file_menu],main_args,main_n);
  XtAddCallback(mw[f_exit_menu],XmNactivateCallback,File_Exit_Callback,state);
  XtVaSetValues(mw[f_exit_menu],XmNmnemonic,'E',NULL);


  /* EDIT MENU */
  mw[edit_menu] = XmCreatePulldownMenu(mw[menu_menu],"editm",main_args,main_n);

  high_n = start_high_n;
  XtSetArg(high_args[high_n],XmNsubMenuId,mw[edit_menu]); high_n++;
  XtSetArg(high_args[high_n],XmNmnemonic,'E'); high_n++;
  mw[e_cascade_menu] = XtCreateManagedWidget(STR_Edit,xmCascadeButtonWidgetClass,mw[menu_menu],high_args,high_n);
  XtAddCallback(mw[e_cascade_menu],XmNhelpCallback,Edit_Help_Callback,state);
  
  mw[e_undo_menu] = XtCreateManagedWidget(STR_Undo,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_undo_menu],XmNactivateCallback,Edit_Undo_Callback,state);
  XtVaSetValues(mw[e_undo_menu],XmNmnemonic,'U',NULL);

  mw[e_redo_menu] = XtCreateManagedWidget(STR_Redo,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_redo_menu],XmNactivateCallback,Edit_Redo_Callback,state);
  XtVaSetValues(mw[e_redo_menu],XmNmnemonic,'R',NULL);

  mw[e_find_menu] = XtCreateManagedWidget(STR_Find,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_find_menu],XmNactivateCallback,Edit_Find_Callback,state);
  XtVaSetValues(mw[e_find_menu],XmNmnemonic,'F',NULL);

  mw[e_select_sep_menu] = XtCreateManagedWidget("",xmSeparatorWidgetClass,mw[edit_menu],sep_args,j);

  mw[e_cut_menu] = XtCreateManagedWidget(STR_Delete_Selection,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_cut_menu],XmNactivateCallback,Edit_Cut_Callback,state);
  XtVaSetValues(mw[e_cut_menu],XmNmnemonic,'C',NULL);

  mw[e_paste_menu] = XtCreateManagedWidget(STR_Insert_Selection,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_paste_menu],XmNactivateCallback,Edit_Paste_Callback,state);
  XtVaSetValues(mw[e_paste_menu],XmNmnemonic,'P',NULL);

  mw[e_mix_menu] = XtCreateManagedWidget(STR_Mix_Selection,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_mix_menu],XmNactivateCallback,Edit_Mix_Callback,state);
  XtVaSetValues(mw[e_mix_menu],XmNmnemonic,'M',NULL);

  mw[e_play_menu] = XtCreateManagedWidget(STR_Play_selection,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_play_menu],XmNactivateCallback,Edit_Play_Callback,state);
  XtVaSetValues(mw[e_play_menu],XmNmnemonic,'P',NULL);

  mw[e_save_as_menu] = XtCreateManagedWidget(STR_Save_selection,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_save_as_menu],XmNactivateCallback,Edit_Save_As_Callback,state);
  XtVaSetValues(mw[e_save_as_menu],XmNmnemonic,'S',NULL);

  mw[e_select_all_menu] = XtCreateManagedWidget(STR_Select_all,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_select_all_menu],XmNactivateCallback,Edit_Select_All_Callback,state);

  mw[e_edit_sep_menu] = XtCreateManagedWidget("",xmSeparatorWidgetClass,mw[edit_menu],sep_args,j);

  mw[e_edenv_menu] = XtCreateManagedWidget(STR_Edit_Envelope,xmPushButtonWidgetClass,mw[edit_menu],main_args,main_n);
  XtAddCallback(mw[e_edenv_menu],XmNactivateCallback,Edit_Envelope_Callback,state);
  XtVaSetValues(mw[e_edenv_menu],XmNmnemonic,'E',NULL);

  mw[e_header_menu] = XtCreateManagedWidget(STR_Edit_Header,xmPushButtonWidgetClass,mw[edit_menu],in_args,in_n);
  XtAddCallback(mw[e_header_menu],XmNactivateCallback,Edit_Header_Callback,state);
  XtVaSetValues(mw[e_header_menu],XmNmnemonic,'H',NULL);

#if (XmVERSION == 1)
  mw[e_history_menu] = XtCreateManagedWidget(STR_Show_edit_history,xmPushButtonWidgetClass,mw[edit_menu],main_args,main_n);
  XtAddCallback(mw[e_history_menu],XmNactivateCallback,Edit_History_Callback,state);
#endif


  /* VIEW MENU */
  mw[view_menu] = XmCreatePulldownMenu(mw[menu_menu],"viewm",main_args,main_n);

  high_n = start_high_n;
  XtSetArg(high_args[high_n],XmNsubMenuId,mw[view_menu]); high_n++;
  XtSetArg(high_args[high_n],XmNmnemonic,'V'); high_n++;
  mw[v_cascade_menu] = XtCreateManagedWidget(STR_View,xmCascadeButtonWidgetClass,mw[menu_menu],high_args,high_n);
  XtAddCallback(mw[v_cascade_menu],XmNhelpCallback,View_Help_Callback,state);

  mw[v_ctrls_menu] = XtCreateManagedWidget(STR_Show_controls,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_ctrls_menu],XmNactivateCallback,View_Ctrls_Callback,state);
  XtVaSetValues(mw[v_ctrls_menu],XmNmnemonic,'S',NULL);

  mw[v_normalize_menu] = XtCreateManagedWidget(STR_Normalize,xmPushButtonWidgetClass,mw[view_menu],in_args,in_n);
  XtAddCallback(mw[v_normalize_menu],XmNactivateCallback,View_Normalize_Callback,state);
  XtVaSetValues(mw[v_normalize_menu],XmNmnemonic,'N',NULL);

  mw[v_listener_menu] = XtCreateManagedWidget(STR_Open_listener,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_listener_menu],XmNactivateCallback,View_Listener_Callback,state);
  XtVaSetValues(mw[v_listener_menu],XmNmnemonic,'L',NULL);

  mw[v_combine_menu] = XmCreatePulldownMenu(mw[view_menu],"combine",main_args,main_n);

  k=main_n;
  XtSetArg(main_args[k],XmNsubMenuId,mw[v_combine_menu]); k++;
  mw[v_combine_cascade_menu] = XtCreateManagedWidget(STR_Channel_style,xmCascadeButtonWidgetClass,mw[view_menu],main_args,k);

  mw[v_combine_separate_menu] = XtCreateManagedWidget(STR_separate,xmPushButtonWidgetClass,mw[v_combine_menu],main_args,main_n);
  XtAddCallback(mw[v_combine_separate_menu],XmNactivateCallback,View_Separate_Callback,state); 
  if (channel_style(state) == CHANNELS_SEPARATE) XtSetSensitive(mw[v_combine_separate_menu],FALSE);

  mw[v_combine_combined_menu] = XtCreateManagedWidget(STR_combined,xmPushButtonWidgetClass,mw[v_combine_menu],main_args,main_n);
  XtAddCallback(mw[v_combine_combined_menu],XmNactivateCallback,View_Combined_Callback,state);  
  if (channel_style(state) == CHANNELS_COMBINED) XtSetSensitive(mw[v_combine_combined_menu],FALSE);

  mw[v_combine_superimposed_menu] = XtCreateManagedWidget(STR_superimposed,xmPushButtonWidgetClass,mw[v_combine_menu],main_args,main_n);
  XtAddCallback(mw[v_combine_superimposed_menu],XmNactivateCallback,View_Superimposed_Callback,state);  
  if (channel_style(state) == CHANNELS_SUPERIMPOSED) XtSetSensitive(mw[v_combine_superimposed_menu],FALSE);


  mw[v_graph_style_menu] = XmCreatePulldownMenu(mw[view_menu],"graph-style",main_args,main_n);

  k=main_n;
  XtSetArg(main_args[k],XmNsubMenuId,mw[v_graph_style_menu]); k++;
  mw[v_graph_style_cascade_menu] = XtCreateManagedWidget(STR_Graph_style,xmCascadeButtonWidgetClass,mw[view_menu],main_args,k);

  mw[v_lines_menu] = XtCreateManagedWidget(STR_lines,xmPushButtonWidgetClass,mw[v_graph_style_menu],main_args,main_n);
  XtAddCallback(mw[v_lines_menu],XmNactivateCallback,View_Lines_Callback,state); 
  if (graph_style(state) == GRAPH_LINES) XtSetSensitive(mw[v_lines_menu],FALSE);

  mw[v_dots_menu] = XtCreateManagedWidget(STR_dots,xmPushButtonWidgetClass,mw[v_graph_style_menu],main_args,main_n);
  XtAddCallback(mw[v_dots_menu],XmNactivateCallback,View_Dots_Callback,state);  
  if (graph_style(state) == GRAPH_DOTS) XtSetSensitive(mw[v_dots_menu],FALSE);

  mw[v_filled_menu] = XtCreateManagedWidget(STR_filled,xmPushButtonWidgetClass,mw[v_graph_style_menu],main_args,main_n);
  XtAddCallback(mw[v_filled_menu],XmNactivateCallback,View_Filled_Callback,state);  
  if (graph_style(state) == GRAPH_FILLED) XtSetSensitive(mw[v_filled_menu],FALSE);

  mw[v_dots_and_lines_menu] = XtCreateManagedWidget(STR_dots_and_lines,xmPushButtonWidgetClass,mw[v_graph_style_menu],main_args,main_n);
  XtAddCallback(mw[v_dots_and_lines_menu],XmNactivateCallback,View_Dots_and_Lines_Callback,state);  
  if (graph_style(state) == GRAPH_DOTS_AND_LINES) XtSetSensitive(mw[v_dots_and_lines_menu],FALSE);

  mw[v_lollipops_menu] = XtCreateManagedWidget(STR_lollipops,xmPushButtonWidgetClass,mw[v_graph_style_menu],main_args,main_n);
  XtAddCallback(mw[v_lollipops_menu],XmNactivateCallback,View_Lollipops_Callback,state);  
  if (graph_style(state) == GRAPH_LOLLIPOPS) XtSetSensitive(mw[v_lollipops_menu],FALSE);


  mw[v_cursor_menu] = XtCreateManagedWidget(STR_Verbose_cursor,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_cursor_menu],XmNactivateCallback,View_Cursor_Callback,state);

  mw[v_sep1_menu] = XtCreateManagedWidget("",xmSeparatorWidgetClass,mw[view_menu],sep_args,j);

  mw[v_region_menu] = XtCreateManagedWidget(STR_Regions,xmPushButtonWidgetClass,mw[view_menu],in_args,in_n);
  XtAddCallback(mw[v_region_menu],XmNactivateCallback,View_Region_Callback,state);
  XtVaSetValues(mw[v_region_menu],XmNmnemonic,'R',NULL);

  mw[v_files_menu] = XtCreateManagedWidget(STR_Files,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_files_menu],XmNactivateCallback,View_Files_Callback,state);
  XtVaSetValues(mw[v_files_menu],XmNmnemonic,'F',NULL);

  mw[v_color_menu] = XtCreateManagedWidget(STR_Color,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_color_menu],XmNactivateCallback,View_Color_Callback,state);

  mw[v_orientation_menu] = XtCreateManagedWidget(STR_Orientation,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_orientation_menu],XmNactivateCallback,View_Orientation_Callback,state);
  XtVaSetValues(mw[v_orientation_menu],XmNmnemonic,'O',NULL);

  mw[v_groups_menu] = XtCreateManagedWidget(STR_Groups,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_groups_menu],XmNactivateCallback,View_Groups_Callback,state);

  mw[v_sep2_menu] = XtCreateManagedWidget("",xmSeparatorWidgetClass,mw[view_menu],sep_args,j);

  mw[v_marks_menu] = XtCreateManagedWidget((show_marks(state)) ? STR_Hide_marks : STR_Show_marks,
					   xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_marks_menu],XmNactivateCallback,View_Marks_Callback,state);
  XtVaSetValues(mw[v_marks_menu],XmNmnemonic,'m',NULL);

  mw[v_zero_menu] = XtCreateManagedWidget(STR_Show_Y0,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_zero_menu],XmNactivateCallback,View_Zero_Callback,state);
  XtVaSetValues(mw[v_zero_menu],XmNmnemonic,'y',NULL);

  mw[v_axes_menu] = XtCreateManagedWidget(STR_Hide_axes,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_axes_menu],XmNactivateCallback,View_Axes_Callback,state);

  mw[v_consoles_menu] = XtCreateManagedWidget(STR_Hide_consoles,xmPushButtonWidgetClass,mw[view_menu],in_args,in_n);
  XtAddCallback(mw[v_consoles_menu],XmNactivateCallback,View_Consoles_Callback,state);

  mw[v_x_axis_menu] = XmCreatePulldownMenu(mw[view_menu],"xaxis",main_args,main_n);

  k=main_n;
  XtSetArg(main_args[k],XmNsubMenuId,mw[v_x_axis_menu]); k++;
  mw[v_x_axis_cascade_menu] = XtCreateManagedWidget(STR_X_axis_units,xmCascadeButtonWidgetClass,mw[view_menu],main_args,k);

  mw[v_x_axis_seconds_menu] = XtCreateManagedWidget(STR_seconds,xmPushButtonWidgetClass,mw[v_x_axis_menu],main_args,main_n);
  XtAddCallback(mw[v_x_axis_seconds_menu],XmNactivateCallback,Options_X_Axis_Seconds_Callback,state);  
  XtSetSensitive(mw[v_x_axis_seconds_menu],FALSE);

  mw[v_x_axis_samples_menu] = XtCreateManagedWidget(STR_samples,xmPushButtonWidgetClass,mw[v_x_axis_menu],main_args,main_n);
  XtAddCallback(mw[v_x_axis_samples_menu],XmNactivateCallback,Options_X_Axis_Samples_Callback,state);  

  mw[v_x_axis_percentage_menu] = XtCreateManagedWidget(STR_percentage,xmPushButtonWidgetClass,mw[v_x_axis_menu],main_args,main_n);
  XtAddCallback(mw[v_x_axis_percentage_menu],XmNactivateCallback,Options_X_Axis_Percentage_Callback,state);  

  mw[v_error_history_menu] = XtCreateManagedWidget(STR_Error_History,xmPushButtonWidgetClass,mw[view_menu],main_args,main_n);
  XtAddCallback(mw[v_error_history_menu],XmNactivateCallback,View_Error_History_Callback,state);



  /* OPTIONS MENU */
  mw[option_menu] = XmCreatePulldownMenu(mw[menu_menu],"optionm",main_args,main_n);

  high_n = start_high_n;
  XtSetArg(high_args[high_n],XmNsubMenuId,mw[option_menu]); high_n++;
  XtSetArg(high_args[high_n],XmNmnemonic,'O'); high_n++;
  mw[o_cascade_menu] = XtCreateManagedWidget(STR_Options,xmCascadeButtonWidgetClass,mw[menu_menu],high_args,high_n);
  XtAddCallback(mw[o_cascade_menu],XmNhelpCallback,Options_Help_Callback,state);

  mw[o_transform_menu] = XtCreateManagedWidget(STR_Transform_Options,xmPushButtonWidgetClass,mw[option_menu],main_args,main_n);
  XtAddCallback(mw[o_transform_menu],XmNactivateCallback,Options_Transform_Callback,state);
  XtVaSetValues(mw[o_transform_menu],XmNmnemonic,'t',NULL);


  mw[o_speed_menu] = XmCreatePulldownMenu(mw[option_menu],"speedstyle",main_args,main_n);

  k=main_n;
  XtSetArg(main_args[k],XmNsubMenuId,mw[o_speed_menu]); k++;
  mw[o_speed_cascade_menu] = XtCreateManagedWidget(STR_Speed_style,xmCascadeButtonWidgetClass,mw[option_menu],main_args,k);

  mw[o_speed_float_menu] = XtCreateManagedWidget(STR_float,xmPushButtonWidgetClass,mw[o_speed_menu],in_args,in_n);
  XtAddCallback(mw[o_speed_float_menu],XmNactivateCallback,Options_Speed_Float_Callback,state);  

  mw[o_speed_semitone_menu] = XtCreateManagedWidget(STR_semitone,xmPushButtonWidgetClass,mw[o_speed_menu],main_args,main_n);
  XtAddCallback(mw[o_speed_semitone_menu],XmNactivateCallback,Options_Speed_Semitone_Callback,state);  

  mw[o_speed_ratio_menu] = XtCreateManagedWidget(STR_ratio,xmPushButtonWidgetClass,mw[o_speed_menu],main_args,main_n);
  XtAddCallback(mw[o_speed_ratio_menu],XmNactivateCallback,Options_Speed_Ratio_Callback,state);  


  mw[o_focus_style_menu] = XmCreatePulldownMenu(mw[option_menu],"focusstyle",main_args,main_n);

  k=main_n;
  XtSetArg(main_args[k],XmNsubMenuId,mw[o_focus_style_menu]); k++;
  mw[o_focus_cascade_menu] = XtCreateManagedWidget(STR_Focus_style,xmCascadeButtonWidgetClass,mw[option_menu],main_args,k);

  mw[o_focus_left_menu] = XtCreateManagedWidget(STR_focus_left,xmPushButtonWidgetClass,mw[o_focus_style_menu],main_args,main_n);
  XtAddCallback(mw[o_focus_left_menu],XmNactivateCallback,Options_Focus_Left_Callback,state);  

  mw[o_focus_right_menu] = XtCreateManagedWidget(STR_focus_right,xmPushButtonWidgetClass,mw[o_focus_style_menu],main_args,main_n);
  XtAddCallback(mw[o_focus_right_menu],XmNactivateCallback,Options_Focus_Right_Callback,state);  

  mw[o_focus_middle_menu] = XtCreateManagedWidget(STR_focus_middle,xmPushButtonWidgetClass,mw[o_focus_style_menu],main_args,main_n);
  XtAddCallback(mw[o_focus_middle_menu],XmNactivateCallback,Options_Focus_Middle_Callback,state);  

  mw[o_focus_active_menu] = XtCreateManagedWidget(STR_focus_active,xmPushButtonWidgetClass,mw[o_focus_style_menu],main_args,main_n);
  XtAddCallback(mw[o_focus_active_menu],XmNactivateCallback,Options_Focus_Active_Callback,state);  
  activate_focus_menu(state,zoom_focus_style(state));

  mw[o_save_menu] = XtCreateManagedWidget(STR_Save_options,xmPushButtonWidgetClass,mw[option_menu],main_args,main_n);
  XtAddCallback(mw[o_save_menu],XmNactivateCallback,Options_Save_Callback,state);
  XtVaSetValues(mw[o_save_menu],XmNmnemonic,'a',NULL);

  mw[o_save_state_menu] = XtCreateManagedWidget(STR_Save_state,xmPushButtonWidgetClass,mw[option_menu],main_args,main_n);
  XtAddCallback(mw[o_save_state_menu],XmNactivateCallback,Options_Save_State_Callback,state);

  mw[o_stats_menu] = XtCreateManagedWidget(STR_Show_stats,xmPushButtonWidgetClass,mw[option_menu],main_args,main_n);
  XtAddCallback(mw[o_stats_menu],XmNactivateCallback,Options_Stats_Callback,state);



  /* HELP MENU */
  mw[help_menu] = XmCreatePulldownMenu(mw[menu_menu],"helpm",main_args,main_n);

  high_n = start_high_n;
  XtSetArg(high_args[high_n],XmNsubMenuId,mw[help_menu]); high_n++;
  XtSetArg(high_args[high_n],XmNmnemonic,'H'); high_n++;
  mw[h_cascade_menu] = XtCreateManagedWidget(STR_Help,xmCascadeButtonWidgetClass,mw[menu_menu],high_args,high_n);
  XtAddCallback(mw[h_cascade_menu],XmNhelpCallback,Help_Help_Callback,state);

  mw[h_click_for_help_menu] = XtCreateManagedWidget(STR_Click_for_help,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_click_for_help_menu],XmNactivateCallback,Help_Context_Help_Callback,state);
  XtVaSetValues(mw[h_click_for_help_menu],XmNmnemonic,'C',NULL);

  mw[h_about_snd_menu] = XtCreateManagedWidget(STR_Overview,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_about_snd_menu],XmNactivateCallback,Help_About_Snd_Callback,state);
  XtVaSetValues(mw[h_about_snd_menu],XmNmnemonic,'O',NULL);

  mw[h_fft_menu] = XtCreateManagedWidget(STR_FFT,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_fft_menu],XmNactivateCallback,Help_FFT_Callback,state);

  mw[h_find_menu] = XtCreateManagedWidget(STR_Find,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_find_menu],XmNactivateCallback,Help_Find_Callback,state);

  mw[h_undo_menu] = XtCreateManagedWidget(STR_Undo_and_redo,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_undo_menu],XmNactivateCallback,Help_Undo_Callback,state);

  mw[h_sync_menu] = XtCreateManagedWidget(STR_Sync,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_sync_menu],XmNactivateCallback,Help_Sync_Callback,state);

  mw[h_speed_menu] = XtCreateManagedWidget(STR_Speed,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_speed_menu],XmNactivateCallback,Help_Speed_Callback,state);

  mw[h_expand_menu] = XtCreateManagedWidget(STR_Expand,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_expand_menu],XmNactivateCallback,Help_Expand_Callback,state);

  mw[h_reverb_menu] = XtCreateManagedWidget(STR_Reverb,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_reverb_menu],XmNactivateCallback,Help_Reverb_Callback,state);

  mw[h_contrast_menu] = XtCreateManagedWidget(STR_Contrast,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_contrast_menu],XmNactivateCallback,Help_Contrast_Callback,state);

  mw[h_env_menu] = XtCreateManagedWidget(STR_Envelope,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_env_menu],XmNactivateCallback,Help_Env_Callback,state);

  mw[h_marks_menu] = XtCreateManagedWidget(STR_Marks,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_marks_menu],XmNactivateCallback,Help_Marks_Callback,state);

  mw[h_mix_menu] = XtCreateManagedWidget(STR_Mixing,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_mix_menu],XmNactivateCallback,Help_Mix_Callback,state);

  mw[h_sound_files_menu] = XtCreateManagedWidget(STR_Formats,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_sound_files_menu],XmNactivateCallback,Help_Sound_Files_Callback,state);

  mw[h_init_file_menu] = XtCreateManagedWidget(STR_Customization,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_init_file_menu],XmNactivateCallback,Help_Init_File_Callback,state);

  mw[h_recording_menu] = XtCreateManagedWidget(STR_Recording,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_recording_menu],XmNactivateCallback,Help_Recording_Callback,state);

#if WITH_MUS_MODULE
  mw[h_clm_menu] = XtCreateManagedWidget(STR_CLM,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_clm_menu],XmNactivateCallback,Help_CLM_Callback,state);
#endif

  mw[h_news_menu] = XtCreateManagedWidget(STR_News,xmPushButtonWidgetClass,mw[help_menu],main_args,main_n);
  XtAddCallback(mw[h_news_menu],XmNactivateCallback,Help_News_Callback,state);

  XtVaSetValues(mw[menu_menu],XmNmenuHelpWidget,mw[h_cascade_menu],NULL);
#ifndef SND_AS_WIDGET
  XtManageChild(mw[menu_menu]);
#endif

  return(mw[menu_menu]);
}

static void GH_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  int callb;
  XtVaGetValues(w,XmNuserData,&callb,NULL);
  g_snd_callback((snd_state *)clientData,callb);
}

#define MAX_MAIN_MENUS 12
static Widget added_menus[MAX_MAIN_MENUS];
static int new_menu = 4;
static Widget *added_options = NULL;
static char **added_options_names = NULL;
static int *added_options_menus = NULL;
static int added_options_size = 0;
static int added_options_pos = 0;

static void add_option(Widget w,int which_menu,char *label)
{
  int i;
  if (added_options_pos == added_options_size)
    {
      added_options_size += 8;
      if (added_options_pos == 0)
	{
	  added_options = (Widget *)CALLOC(added_options_size,sizeof(Widget));
	  added_options_names = (char **)CALLOC(added_options_size,sizeof(char *));
	  added_options_menus = (int *)CALLOC(added_options_size,sizeof(int));
	}
      else
	{
	  added_options = (Widget *)REALLOC(added_options,added_options_size * sizeof(Widget));
	  added_options_names = (char **)REALLOC(added_options_names,added_options_size * sizeof(char *));
	  added_options_menus = (int *)REALLOC(added_options_menus,added_options_size * sizeof(int));
	  for (i=added_options_pos;i<added_options_size;i++) added_options[i] = NULL;
	}
    }
  added_options[added_options_pos] = w;
  added_options_menus[added_options_pos] = which_menu;
  added_options_names[added_options_pos] = copy_string(label);
  added_options_pos++;
}

static int remove_option(int which_menu,char *label)
{
  int i;
  for (i=0;i<added_options_pos;i++)
    {
      if ((added_options_menus[i] == which_menu) && (strcmp(label,added_options_names[i]) == 0) && (added_options[i]))
	{
	  XtDestroyWidget(added_options[i]);
	  added_options[i] = NULL;
	  added_options_menus[i] = -1;
	  FREE(added_options_names[i]);
	  added_options_names[i] = NULL;
	  return(0);
	}
    }
  return(-1);
}

int change_menu_label(int which_menu,char *old_label, char *new_label)
{
  int i;
  for (i=0;i<added_options_pos;i++)
    {
      if ((added_options_menus[i] == which_menu) && (strcmp(old_label,added_options_names[i]) == 0) && (added_options[i]))
	{
	  make_name_label(added_options[i],new_label);
	  if (added_options_names[i]) FREE(added_options_names[i]);
	  added_options_names[i] = copy_string(new_label);
	  return(0);
	}
    }
  return(-1);
}

int set_menu_sensitive(int which_menu,char *old_label, int on)
{
  int i;
  for (i=0;i<added_options_pos;i++)
    {
      if ((added_options_menus[i] == which_menu) && (strcmp(old_label,added_options_names[i]) == 0) && (added_options[i]))
	{
	  XtSetSensitive(added_options[i],on);
	  return(0);
	}
    }
  return(-1);
}

int add_to_main_menu(snd_state *ss, char *label)
{
  static Arg args[12];
  Widget m;
  int n;
  n=0;
  if (!(ss->using_schemes)) n = background_basic_color(args,n,ss);
  m = XmCreatePulldownMenu(mw[menu_menu],label,args,n);

  n=0;
  if (!(ss->using_schemes)) n = background_highlight_color(args,n,ss);
  XtSetArg(args[n],XmNsubMenuId,m); n++;
  XtCreateManagedWidget(label,xmCascadeButtonWidgetClass,mw[menu_menu],args,n);

  new_menu++;
  if (new_menu < MAX_MAIN_MENUS)
    {
      added_menus[new_menu] = m;
      return(new_menu);
    }
  else
    {
      ss->eval_error = SND_NO_SUCH_MENU;
      return(MAX_MAIN_MENUS - 1);
    }
}

void add_to_menu(snd_state *ss, int which_menu, char *label, int callb)
{
  Widget m,menw;
  static Arg args[12];
  int n;
  switch (which_menu)
    {
    case 0: menw = mw[file_menu]; break;
    case 1: menw = mw[edit_menu]; break;
    case 2: menw = mw[view_menu]; break;
    case 3: menw = mw[option_menu]; break;
    case 4: menw = mw[help_menu]; break;
    default: 
      if (which_menu < MAX_MAIN_MENUS)
	menw = added_menus[which_menu]; 
      else 
	{
	  ss->eval_error = SND_NO_SUCH_MENU;
	  return;
	}
      break;
    }
  n=0;
  if (!(ss->using_schemes)) n = background_basic_color(args,n,ss);
  XtSetArg(args[n],XmNuserData,callb); n++;
  m = XtCreateManagedWidget(label,xmPushButtonWidgetClass,menw,args,n);
  XtAddCallback(m,XmNactivateCallback,GH_Callback,ss);
  add_option(m,which_menu,label);
}

int remove_from_menu(snd_state *ss, int which_menu, char *label)
{
  return(remove_option(which_menu,label));
}



/* -------------------------------- POPUP MENU -------------------------------- */

static void Popup_Play_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  snd_info *sp;
  start_playing(sp=any_selected_sound(ss),0);
  set_play_button(sp,1);
}

static void Popup_Save_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  save_edits(any_selected_sound(ss),NULL);
}

static void Popup_Undo_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  undo_EDIT((void *)clientData,1);
}

static void Popup_Redo_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  redo_EDIT((void *)clientData,1);
}

static void Popup_Normalize_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  normalize_all_sounds((snd_state *)clientData);
}

static void Popup_Info_Callback(Widget w,XtPointer clientData,XtPointer callData) 
{
  snd_state *ss = (snd_state *)clientData;
  snd_info *sp;
  sp = selected_sound(ss);
  if (sp) display_info(sp);
}

static void Post_Popup_Menu(Widget w,XtPointer clientData,XEvent *event,Boolean *flag)
{
  if (event->xbutton.button == Button3)
    {
      XmMenuPosition(popup_menu,(XButtonPressedEvent *)event);
      XtManageChild(popup_menu);
    }
}

void create_popup_menu(snd_state *ss)
{
  /* make it a child of the main window */
  Widget mainp;
  Arg args[20];
  int n;
  if (!popup_menu)
    {
      n=0;
      if (!ss->using_schemes) n = background_basic_color(args,n,ss); else n=0;
      mainp = main_PANE(ss);
      popup_menu = XmCreatePopupMenu(mainp,"popup-menu",args,n);
      XtAddEventHandler(mainp,ButtonPressMask,FALSE,Post_Popup_Menu,popup_menu);

      popup_children[W_pop_menu] = XtCreateManagedWidget("snd",xmLabelWidgetClass,popup_menu,args,n);
      popup_children[W_pop_sep] = XtCreateManagedWidget("sep",xmSeparatorWidgetClass,popup_menu,args,n);
      popup_children[W_pop_play] = XtCreateManagedWidget(STR_Play,xmPushButtonWidgetClass,popup_menu,args,n);
      XtAddCallback(popup_children[W_pop_play],XmNactivateCallback,Popup_Play_Callback,ss);
      popup_children[W_pop_undo] = XtCreateManagedWidget(STR_Undo,xmPushButtonWidgetClass,popup_menu,args,n);
      XtVaSetValues(popup_children[W_pop_undo],XmNsensitive,FALSE,NULL);
      XtAddCallback(popup_children[W_pop_undo],XmNactivateCallback,Popup_Undo_Callback,ss);
      popup_children[W_pop_redo] = XtCreateManagedWidget(STR_Redo,xmPushButtonWidgetClass,popup_menu,args,n);
      XtVaSetValues(popup_children[W_pop_redo],XmNsensitive,FALSE,NULL);
      XtAddCallback(popup_children[W_pop_redo],XmNactivateCallback,Popup_Redo_Callback,ss);
      popup_children[W_pop_save] = XtCreateManagedWidget(STR_Save,xmPushButtonWidgetClass,popup_menu,args,n);
      XtAddCallback(popup_children[W_pop_save],XmNactivateCallback,Popup_Save_Callback,ss);
      popup_children[W_pop_normalize] = XtCreateManagedWidget(STR_Normalize,xmPushButtonWidgetClass,popup_menu,args,n);
      XtVaSetValues(popup_children[W_pop_normalize],XmNsensitive,FALSE,NULL);
      XtAddCallback(popup_children[W_pop_normalize],XmNactivateCallback,Popup_Normalize_Callback,ss);
      popup_children[W_pop_info] = XtCreateManagedWidget(STR_Info,xmPushButtonWidgetClass,popup_menu,args,n);
      XtVaSetValues(popup_children[W_pop_info],XmNsensitive,FALSE,NULL);
      XtAddCallback(popup_children[W_pop_info],XmNactivateCallback,Popup_Info_Callback,ss);
    }
}

void add_popup_handler(Widget w)
{
  XtAddEventHandler(w,ButtonPressMask,FALSE,Post_Popup_Menu,popup_menu);
}
